pub struct MemoryStore {
tickets: HashMap<ChannelId, MemoryTicketQueue>,
out_indices: HashMap<(ChannelId, u32), u64>,
}Expand description
Simple non-persistent ticket queue store backed by a HashMap and MemoryTicketQueue.
Useful for non-persistent and testing scenarios.
Fields
tickets: HashMap<ChannelId, MemoryTicketQueue>out_indices: HashMap<(ChannelId, u32), u64>Trait Implementations
Sourceimpl Clone for MemoryStore
impl Clone for MemoryStore
Sourcefn clone(&self) -> MemoryStore
fn clone(&self) -> MemoryStore
Returns a duplicate of the value. Read more
1.0.0 · Sourcefn clone_from(&mut self, source: &Self)
fn clone_from(&mut self, source: &Self)
Performs copy-assignment from
source. Read moreSourceimpl Debug for MemoryStore
impl Debug for MemoryStore
Sourceimpl Default for MemoryStore
impl Default for MemoryStore
Sourcefn default() -> MemoryStore
fn default() -> MemoryStore
Returns the “default value” for a type. Read more
Sourceimpl OutgoingIndexStore for MemoryStore
impl OutgoingIndexStore for MemoryStore
type Error = Infallible
Sourcefn load_outgoing_index(
&self,
channel_id: &ChannelId,
epoch: u32,
) -> Result<Option<u64>, Self::Error>
fn load_outgoing_index( &self, channel_id: &ChannelId, epoch: u32, ) -> Result<Option<u64>, Self::Error>
Loads the last used outgoing ticket index for the given channel and epoch. Read more
Sourcefn save_outgoing_index(
&mut self,
channel_id: &ChannelId,
epoch: u32,
index: u64,
) -> Result<(), Self::Error>
fn save_outgoing_index( &mut self, channel_id: &ChannelId, epoch: u32, index: u64, ) -> Result<(), Self::Error>
Saves the last used outgoing ticket index for the given channel and epoch.
Sourceimpl TicketQueueStore for MemoryStore
impl TicketQueueStore for MemoryStore
Sourcetype Queue = MemoryTicketQueue
type Queue = MemoryTicketQueue
Type of per-channel incoming ticket queues.
Sourcefn open_or_create_queue(
&mut self,
channel_id: &ChannelId,
) -> Result<Self::Queue, <Self::Queue as TicketQueue>::Error>
fn open_or_create_queue( &mut self, channel_id: &ChannelId, ) -> Result<Self::Queue, <Self::Queue as TicketQueue>::Error>
Opens or creates a new queue in storage for the given channel.
Sourcefn delete_queue(
&mut self,
channel_id: &ChannelId,
) -> Result<Vec<VerifiedTicket>, <Self::Queue as TicketQueue>::Error>
fn delete_queue( &mut self, channel_id: &ChannelId, ) -> Result<Vec<VerifiedTicket>, <Self::Queue as TicketQueue>::Error>
Deletes the queue for the given channel. Read more
Sourcefn iter_queues(
&self,
) -> Result<impl Iterator<Item = ChannelId>, <Self::Queue as TicketQueue>::Error>
fn iter_queues( &self, ) -> Result<impl Iterator<Item = ChannelId>, <Self::Queue as TicketQueue>::Error>
Iterate over all channel IDs of ticket queues in the storage.