RedbStore

Struct RedbStore 

pub struct RedbStore {
    db: Arc<Database>,
    _tmp: Option<NamedTempFile>,
}
Implementation of OutgoingIndexStore and TicketQueueStore using redb database and postcard serializer.

The store is intentionally not cloneable to allow the owner to have complete ownership of the database.

Fields

db: Arc<Database>_tmp: Option<NamedTempFile>

Implementations

Source

impl RedbStore

pub fn new(path: impl AsRef<Path>) -> Result<Self, RedbStoreError>

Creates a new instance on the given path.

pub fn new_temp() -> Result<Self, RedbStoreError>

Creates a new instance using a temporary file.

The temporary file is automatically deleted when the store is dropped.